//二叉堆（最大堆）的插入和删除算法的验证
//  要点：
//      1 二叉堆性质的描述
//      2 \old 的使用

#include <stdlib.h>

#define MAX_ELEMENTS 1000
#define HEAP_FULL(n)  (n == MAX_ELEMENTS - 1)
#define HEAP_EMPTY(n) (n == 0)
typedef struct{               // 节点的数据类型
            int key;
            /* other fields */
        } Element;
Element bheap[MAX_ELEMENTS]={0};  // 储存二叉堆的数组

/////////////////////////////////////insert_max_heap函数的验证//////////////////////////////////////
/*@
    requires 
        \old(*n) == *n && MAX_ELEMENTS - 1 >= *n >= 0             // *n = 二叉堆的长度
        && (\forall int i:[2..*n].bheap[i].key <= bheap[i/2].key) // 最大二叉堆的基本性质
        ;
    assigns bheap, *n;
    ensures 
        *n == \old(*n) + 1 
        && (\forall int i:[2..*n].bheap[i].key <= bheap[i/2].key) // 插入后最大二叉堆保持性质不变
        ;
    exits 
        \exit_status == 1  // 超出数组范围时的退出状态
        ;
*/
void insert_max_heap(Element item, int * const n)
{
    int i;
    if(HEAP_FULL(*n)){  // 堆空间（数组）已满
        exit(1);
    }
    i = ++(*n);  // 新节点插入的位置
    /*@ loop invariant
        i == *n && \old(*n) >= 0 && *n == \old(*n) + 1  
        && (\forall int j:[2..\old(*n)].bheap[j].key <= bheap[j/2].key)
        && MAX_ELEMENTS - 1 > \old(*n) >= 0
        ||
        i <= *n / 2 &&  i > 0 && item.key > bheap[i].key && *n == \old(*n) + 1 
        && (\forall int j:[2..*n].bheap[j].key <= bheap[j/2].key) 
        && MAX_ELEMENTS - 1 > \old(*n) >= 0
        ;
        loop variant i-1; 
    */
    while((i != 1) && (item.key > bheap[i/2].key)){
          bheap[i].key = bheap[i/2].key;
        i /= 2;
    }
    bheap[i] = item;
}


/////////////////////////////////////delete_max_heap函数的验证//////////////////////////////////////
/*@
    requires
        \old(*n) == *n && MAX_ELEMENTS - 1 >= *n >= 0 
        && (\forall int j:[2..*n].bheap[j].key <= bheap[j/2].key) 
        && \old(bheap[1]) == bheap[1]  //利用\old储存函数入口处二叉堆的最大元素
            
        ;
    assigns bheap, *n;
    ensures 
        \old(*n) != 0 && *n == \old(*n) -1 && \result == \old(bheap[1])
        && (\forall int j:[2..*n].bheap[j].key <= bheap[j/2].key) 
            //删除元素后，最大二叉堆性质保持不变，长度减1
        ;
    exits 
        \exit_status == 1
        ;
*/
Element delete_max_heap(int * const n)    // 删除并返回最大二叉堆的根节点（最大元素）
{
    int parent, child;
    Element maxItem, lastItem;

    if(HEAP_EMPTY(*n)){    // 堆空间（数组）已空
        exit(1);        
    }
    maxItem = bheap[1];
    lastItem = bheap[(*n)--];
    parent = 1;
    child = 2;
    /*@ loop invariant
        *n == \old(*n) - 1 && MAX_ELEMENTS - 1 > *n >= 0 && maxItem == \old(bheap[1]) 
        && child == parent * 2 && (\forall int j:[2..*n].bheap[j].key <= bheap[j/2].key) 
        &&
        (parent == 1 && child == 2 && lastItem.key <= bheap[parent].key 
             //循环入口
         ||
         parent > 1  && parent <= *n && child < *n && bheap[child].key >= bheap[child + 1].key 
         && lastItem.key < bheap[parent].key 
            //循环中根据if条件得到的分支1
         ||
         parent > 1  && parent <= *n && child + 1 <= *n && bheap[child].key < bheap[child + 1].key 
         && lastItem.key < bheap[parent].key 
             //循环中根据if条件得到的分支2
         ||
         parent > 1  && parent <= *n && child >= *n 
         && lastItem.key < bheap[parent].key  
             //循环中根据if条件得到的分支3
        )
        ;
        loop variant *n+1 - child;
    */
    while(child <= *n){
        if ((child < *n) && (bheap[child].key < bheap[child +1].key)){  // 存在2个子节点时，取大的
            child++;
        }
        if(lastItem.key >= bheap[child].key){ // 位置调整达成
            break;
        }
        bheap[parent] = bheap[child];
        parent = child;
        child *= 2;        
    }
    bheap[parent] = lastItem;
    return maxItem;    
}